#include <iomanip>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <cmath>
#include <cstdlib>
#include <cuba.h>
#include <gsl/gsl_errno.h>

#include "commons.hh"
#include "math/matrix.hh"
#include "expansion_AD_SF.hh"


using namespace std;



namespace HELL {


  bool a10woRC = false;
  bool fullySymm = false;
  bool shiftMmin = true;



  dcomplex fdamp(dcomplex N) {
    return 1./(N+1.);
  }
  dcomplex fmom(dcomplex N) { // same form as used in preHELL
    //return 12.*N/pow(1.+N,2)/(2.+N);
    //return 8.*N/pow(1.+N,3);
    return 4.*N/pow(N+1.,2.);
  }




  const double c0 = CA/M_PI*4*log(2.);
  const double k0 = CA/M_PI*28.*ZETA3;
  // from DL-LO kernel
  inline double c1LO (int nf) { return -15.00496429 - 0.04503163717*nf;}
  inline double k1LO (int nf) { return -507.744719  - 1.080759292*nf;}
  // from DL-NLO kernel
  //inline double c1NLO(int nf) { return -15.04043249 - 0.2076545137*nf; }
  //inline double k1NLO(int nf) { return -574.496771  - 0.3724353405*nf; }
  //inline double m1NLO(int nf) { return 0.3099881585 - 0.01878716112*nf; }
  inline double c1NLO(int nf) { return -11.696833425 - 0.4102968810*nf; }
  inline double k1NLO(int nf) { return -494.250393369 - 5.23585215538*nf; }
  inline double m1NLO(int nf) { return -0.1492429211 + 0.00904502552*nf; }
  //
  double a11()       { return CA/M_PI; }
  double a10(int nf) { return -(11.*CA + 2.*nf*(1.-2.*CF/CA))/12./M_PI + (a10woRC ? beta0(nf) : 0 ); }
  double a21(int nf) { return nf*(26*CF-23*CA)/36/M_PI/M_PI; }
  double a20(int nf) { return (19.6798969734649 + nf*(4./9.*ZETA2-68./81.) + nf*nf*13./2187.)/M_PI/M_PI; }
  double a32(int nf) { return -CA*( (18.*ZETA2-71.)*(CA-2*CF)*nf + CA*CA*(54*ZETA3+99.*ZETA2-395) ) /108./M_PI/M_PI/M_PI; }
  //
  dcomplex gamma0approx(dcomplex N, double nf) {
    double a1 = a11();
    double a0 = a10(nf) - (a10woRC ? beta0(nf) : 0 );
    return a1/N + a0 + 2*(a1+a0)*(1./(N+1.)-1.); // preserves momentum conservation, accurate at NLL, constant at large N
    //return gamma0pl(N+1., 0) + fdamp(N)*nf*(2*CF/CA-1)/6./M_PI;
  }
  dcomplex gamma0singNoDamp(dcomplex N, double nf) {
    double momcons = 0;
    double a1 = a11();
    double a0 = a10(nf) - (a10woRC ? beta0(nf) : 0 );
    return a1/N + a0*fdamp(N) - momcons*fmom(N)*(a1+a0*fdamp(1.)); // preserves momentum conservation, accurate at NLL, zero at large N
  }
  dcomplex gamma0sing(dcomplex N, double nf) {
    //return gamma0approx(N,nf);  // I'd discard this option, as a purely resummed result should vanish at large N, ideally...
    return gamma0singNoDamp(N,nf);
    //return gamma0singNoDamp(N,nf)/(N+1.);
    //return gamma0singNoDamp(N,nf) - 2.*gamma0singNoDamp(N+1.,nf) + gamma0singNoDamp(N+2.,nf);
  }
  dcomplex gamma1singLLp(dcomplex N, double nf) {
    double c0 = 4*log(2.);
    double k0 = 28*ZETA3;
    return beta0(nf)*CA/M_PI*(3*k0/32.-c0) * (1./N - fmom(N)); // add the correct as^2 contribution (from Bateman only - the DL part is zero)
    //return beta0(nf)*CA/M_PI*(-c0/4.) * (1./N - fmom(N)); // add the correct as^2 contribution (from Bateman only - the DL part is zero)
  }
  dcomplex gamma1singNLL(dcomplex N, double nf) {
    //double momcons = 0;
    double a1 = a21(nf);
    //double a0 = a20(nf);
    //return ( a1/N + a0*fdamp(N) - momcons*fmom(N)*(a1+a0*fdamp(1.)) ); // preserves momentum conservation, accurate at NNLL, zero at large N
    return a1/N - 2*a1/(N+1.); // preserves momentum conservation, accurate at NLL, zero at large N
  }
  dcomplex gamma1sing(dcomplex N, double nf) {
    return ( useLLp ? gamma1singLLp(N,nf) : gamma1singNLL(N,nf) );
  }

dcomplex dpsi(dcomplex Z, int M=1){
  int K1, K2;
  dcomplex SUB = 0. , SUBM;
  dcomplex ZZ = Z;
  if(fabs(imag(ZZ))<10.) { // if too close to the real axis...
  label1:
    SUBM = -1./ZZ;
    for(K1=1; K1<=M; K1++) {
      SUBM = - SUBM * (double)K1 / ZZ;
    }
    if(real(ZZ)<10.) { // ...use recurrence relation to push real(z) large enough
      SUB = SUB + SUBM;
      ZZ = ZZ + 1.;
      goto label1;
    }
  }
  // Expansion coefficients for the first derivative
  double A1 =  1.;
  double A2 =  1./2.;
  double A3 =  1./6.;
  double A4 = -1./30.;
  double A5 =  1./42.;
  double A6 = -1./30.;
  double A7 =  5./66.;
  // Expansion coefficients for the higher derivatives
  if(M>1) {
    for(K2=2;K2<=M;K2++){
      A1 = A1 * (1.*K2-1.);
      A2 = A2 *  1.*K2;
      A3 = A3 * (1.*K2+1.);
      A4 = A4 * (1.*K2+3.);
      A5 = A5 * (1.*K2+5.);
      A6 = A6 * (1.*K2+7.);
      A7 = A7 * (1.*K2+9.);
    }
  }
  dcomplex RZ = 1./ ZZ;
  dcomplex DZ = RZ * RZ;
  // SUB + asympt expansion (Abramowitz, Stengun, 6.4.11)
  return SUB + pow(-1.,M+1) * pow(RZ,M) * ( A1 + RZ * (A2 + RZ * (A3 + DZ * (A4 + DZ * (A5 + DZ * (A6 + A7 * DZ ))))) );
}
  dcomplex gamma_DLNLO_order_as3(dcomplex N, int nf) {
    double a1 = a11(), a0=a10(nf);
    double chi01 = a1;
    double chi02 = a0*a1;
    double chi11 = a21(nf);
    double rho = chi01*chi01*(5*ZETA3/2.-1./24.) + chi02*(53./18.-ZETA2) - chi11;
    //double rho = ( CA*CA*(-74./27. +11./12.*ZETA2 +5./2.*ZETA3) +nf*((4*CA+7*CF)/27. +(CA-2*CF)*ZETA2/6.) )/M_PI/M_PI;
    dcomplex gamma0  =  a1/N + a0 - 2*(a1+a0)*N/(N+1.);
    dcomplex gamma0p = -a1/N/N - 2*(a1+a0)/(N+1.)/(N+1.);
    double b0 = beta0(nf);
    dcomplex res = rho + chi11/(N+1.) + chi02/pow(N+1.,2) - (rho+chi11/2.+chi02/4.-b0*chi01)*fmom(N) + chi01*(dpsi(N+1.)-dpsi(1))*(gamma0+b0);
    if(a10woRC)   res -= b0 * chi01 * (dpsi(N+1.)-dpsi(1) + fmom(N));
    if(fullySymm) res -= b0 * chi01 * (dpsi(N+1.)-dpsi(1) + fmom(N));
    return -gamma0p * res;
  }
  dcomplex gamma_BatemanColl_minus_dcNLO_momcons_order_as3(dcomplex N, int nf) {
    double b0 = beta0(nf);
    double res = b0*k0/16.;
    return b0* res * (1./N - fmom(N));
  }
  dcomplex gammaNLL2(dcomplex N, int nf) {
    return gamma_DLNLO_order_as3(N,nf) + gamma_BatemanColl_minus_dcNLO_momcons_order_as3(N,nf);
  }
  /*
  dcomplex gammaNLL2simpl(dcomplex N, int nf) {
    dcomplex res = 0;
    double a1 = a11(), a0=a10(nf);
    double A = ( CA*CA*(-74./27. +11./12.*ZETA2 +5./2.*ZETA3) +nf*((4*CA+7*CF)/27. +(CA-2*CF)*ZETA2/6.) )/M_PI/M_PI;
    double chi11 = a21(nf);
    double chi02 = a1*a0;
    dcomplex gamma0p = -a1/N/N - 2*(a1+a0)/(N+1.)/(N+1.);
    res  = -gamma0p * ( A + chi11/(N+1.) + chi02/pow(N+1.,2) - (A+chi11/2.+chi02/4.)*fmom(N) );
    res += CA/M_PI* ( a1*a1*  ( 3*ZETA4/N -2*ZETA3/N/N -3./(N+1.) -1./(N+1.)/(N+1.) )
		      +a1*a0* (-2*ZETA3/N +2./(N+1.) +1./(N+1.)/(N+1.) )
		      +4*pow(a0 + a1,2)*              ( 3*ZETA4/(N+1.) -2*ZETA3/(N+1.)/(N+1.) + 1./pow(N+1.,5) )
		      -2*(a0*a0 + 4*a0*a1 + 3*a1*a1)* ( -2*ZETA3/(N+1.) +1./pow(N+1.,4) )
		      );
    // add RC
    double b0 = beta0(nf);
    res += b0*b0*k0/16.* (1./N - fmom(N));
    return res;
  }
  */
  dcomplex gamma2singLLp(dcomplex N, int nf, bool var) {
    dcomplex res = 0;
    double a1 = a11(), a0=a10(nf);
    double chi02 = a1*a0;
    dcomplex gamma0  =  a1/N + a0 - 2*(a1+a0)*N/(N+1.);
    dcomplex gamma0p = -a1/N/N - 2*(a1+a0)/(N+1.)/(N+1.);
    res  = -gamma0p * ( chi02/pow(N+1.,2) - chi02/4.*fmom(N) + CA/M_PI*(dpsi(N+1.)-dpsi(1))*gamma0 );
    // add RC
    double b0 = beta0(nf);
    int T = (var ? 1 : 2);
    double  c1 = c1NLO(nf);
    double  k1 = k1NLO(nf);
    double  m1 = m1NLO(nf);
    double dc1 = c1-c1LO(nf);
    double dk1 = k1-k1LO(nf);
    double dm1 = m1;
    if(shiftMmin) m1 += c0/2.;
    double lam2 = b0*(48*c0*k0-3*k0*k0-256*c0*c0)/256. + (16*c0*(dk1+6*k0*dm1)+k0*(16*dc1-3*dk1-15*k0*dm1))/512.;
    double lam1 = b0*(3*T*k1/32 -T*c1 + (b0+6*m1)*k0/16.);
    res += lam1/N + lam2/N/N - (lam1+lam2)*fmom(N);
    return res;
  }  
  dcomplex gamma2sing(dcomplex N, double nf, bool var) {
    return ( useLLp ? gamma2singLLp(N,nf,var) : gammaNLL2(N,nf) );
    //return ( useLLp ? gamma2singLLp(N,nf) : gammaNLL2simpl(N,nf) );
  }










  double Pole(double x, int k, int j) { // inverse Mellin of 1/N^k/(1+N)^j
    if(j==0 && k>0) return pow(-log(x),k-1)/factorial(k-1)/x;
    if(k==0 && j>0) return pow(-log(x),j-1)/factorial(j-1);
    if(k==1 && j==1) return 1/x-1;
    if(k==2 && j==1) return Pole(x,2,0) - Pole(x,1,0) + Pole(x,0,1);
    if(k==1 && j==2) return Pole(x,1,0) - Pole(x,0,2) - Pole(x,0,1);
    cout << "HELL warning: this inverse mellin is not implemented: 1/N^"<<k<<"/(1+N)^"<<j << endl;
    return 0;
  }
  double exactPoly(double x, int k, int j) { // inverse Mellin of 1/N^k/(1+N)^j (psi_1(1+N)-Zeta2)
    double res = 0;
    if     (k==2 && j==0) res = (2*(Li3(x)-ZETA3) - Li2(x)*log(x))/x;
    else if(k==3 && j==0) res = (3*(ZETA4-Li4(x))+(Li3(x)+2*ZETA3)*log(x))/x;
    else if(k==0 && j==2) res = 2*(Li3(x)-ZETA3) - Li2(x)*log(x) - pow(log(x),3)/6;
    else if(k==0 && j==3) res = 3*(ZETA4-Li4(x))+(Li3(x)+2*ZETA3)*log(x) + pow(log(x),4)/24;
    else cout << "HELL warning: this inverse mellin is not implemented: (psi_1(1+N)-Zeta2)/N^"<<k<<"/(1+N)^"<<j << endl;
    return res;
  }
  //
  double P0(double x, int nf) {
    double a1 = a11(), a0=a10(nf) - (a10woRC ? beta0(nf) : 0 );
    double res = a1/x + a0;
    return res;
  }
  double P0sq(double x, int nf) {
    double a1 = a11(), a0=a10(nf) - (a10woRC ? beta0(nf) : 0 );
    double res = -a1*a1*log(x)/x - a0*a0*log(x) + 2*a1*a0*(1/x-1);
    return res;
  }
  double P0cb(double x, int nf) {
    double a1 = a11(), a0=a10(nf) - (a10woRC ? beta0(nf) : 0 );
    double res = a1*a1*a1*Pole(x,3,0) + a0*a0*a0*Pole(x,0,3) + 3*a1*a1*a0*Pole(x,2,1) + 3*a1*a0*a0*Pole(x,1,2);
    return res;
  }
  double P1(double x, int nf) {
    double res = 0;
    if(useLLp) {
      res = beta0(nf)*(3*k0/32.-c0) * (1/x-4*(1+log(x)));
    } else {
      double a1 = a21(nf);
      res = a1 * (1/x-2);
    }
    return res;
  }
  double P0P1(double x, int nf) {
    double a1 = a11(), a0=a10(nf) - (a10woRC ? beta0(nf) : 0 );
    double res = 0;
    if(useLLp) {
      res = beta0(nf)*(3*k0/32.-c0) * ( a1*(Pole(x,2,0)-4*Pole(x,0,2)) + a0*(Pole(x,1,0)+4*Pole(x,0,3)-4*Pole(x,0,2)-Pole(x,0,1)) );
    } else {
      res = a21(nf) * ( a1*Pole(x,2,0) + (a0-2*a1)/x -2*a0*Pole(x,0,2) + 2*a1-a0 );
    }
    return res;
  }
  double P2(double x, int nf, int var) {
    double res = 0;
    double a1 = a11(), a0=a10(nf);
    double chi02 = a1*a0;
    double b0 = beta0(nf);
    if(useLLp) {
      res = a1*chi02 * ( Pole(x,2,0) - 3*Pole(x,1,0) + 2*Pole(x,0,2) + 3*Pole(x,0,1) )
	+ 2*(a0 + a1)*chi02 * (2*Pole(x,0,4)-Pole(x,0,3));
      res += CA/M_PI* ( a1*a1*  exactPoly(x,3,0) 
			+a1*a0* exactPoly(x,2,0)
			+4*pow(a0 + a1,2)*              exactPoly(x,0,3)
			-2*(a0*a0 + 4*a0*a1 + 3*a1*a1)* exactPoly(x,0,2)
			);
      // add RC
      int T = (var==1 ? 1 : 2);
      double  c1 = c1NLO(nf);
      double  k1 = k1NLO(nf);
      double  m1 = m1NLO(nf);
      double dc1 = c1-c1LO(nf);
      double dk1 = k1-k1LO(nf);
      double dm1 = m1;
      if(shiftMmin) m1 += c0/2.;
      double lam2 = b0*(48*c0*k0-3*k0*k0-256*c0*c0)/256. + (16*c0*(dk1+6*k0*dm1)+k0*(16*dc1-3*dk1-15*k0*dm1))/512.;
      double lam1 = b0*(3*T*k1/32 -T*c1 + (b0+6*m1)*k0/16.);
      res += lam1*Pole(x,1,0) + lam2*Pole(x,2,0) - 4*(1+log(x))*(lam1+lam2);
    } else {
      double chi01 = a1;
      double chi11 = a21(nf);
      double rho = chi01*chi01*(5*ZETA3/2.-1./24.) + chi02*(53./18.-ZETA2) - chi11;
      //double rho = ( CA*CA*(-74./27. +11./12.*ZETA2 +5./2.*ZETA3) +nf*((4*CA+7*CF)/27. +(CA-2*CF)*ZETA2/6.) )/M_PI/M_PI;
      res = (rho*a1 + a1*chi02 + a1*chi11) * Pole(x,2,0)
	+ (-4*rho*a1 - 3*a1*chi02 - 3*a1*chi11) * Pole(x,1,0)
	+ 4*(a0 + a1)*(2*rho + chi02 + chi11) * Pole(x,0,4)
	- 2*(a0 + a1)*(4*rho + chi02 + chi11) * Pole(x,0,3)
	+ 2*(rho*a0 + 3*rho*a1 + a1*chi02 + a1*chi11) * Pole(x,0,2)
	+ (4*rho*a1 + 3*a1*chi02 + 3*a1*chi11) * Pole(x,0,1)
	;
      res += CA/M_PI* ( a1*a1*  exactPoly(x,3,0) 
			+a1*a0* exactPoly(x,2,0)
			+4*pow(a0 + a1,2)*              exactPoly(x,0,3)
			-2*(a0*a0 + 4*a0*a1 + 3*a1*a1)* exactPoly(x,0,2)
			);
      if(!a10woRC && !fullySymm) res += b0 * chi01 * (a1*exactPoly(x,2,0) + 2*(a1+a0)*exactPoly(x,0,2) + a1*4*(Pole(x,1,0)-Pole(x,0,1)-Pole(x,0,2)) +2*(a1+a0)*(Pole(x,0,3)-Pole(x,0,4)) );
      // add RC
      res += b0*b0*k0/16* (1/x-4*(1+log(x)));
    }
    // result
    return res;
  }



};

